<?php

namespace Wildix\Faxes;

use User;
use Wildix\Error;
use Wildix\Exception;
use Wildix\Validator;

/**
 * Class FaxSmsProfiles
 */
class FaxSmsProfiles {
    const COMPANIES_DIRECTORY    = '/etc/companies.d/';
    const PROFILE_TEST_DIRECTORY = '/tmp/session/';

    const OAUTH_CONFIG_ACCESS_FILE_PATTERN  = '/etc/companies.d/0/fetchmail_access_%s';
    const OAUTH_CONFIG_REFRESH_FILE_PATTERN = '/etc/companies.d/0/fetchmail_refresh_%s';

    const OAUTH_ACCESS_EXPIRATION_SECONDS  = 3600;
    const OAUTH_ACCESS_TOKEN_REFRESH_LIMIT = 300;

    const IMAP_PROTOCOL = 'imap';
    const POP3_PROTOCOL = 'pop3';

    const OAUTH_PROVIDER_PROTOCOL_SERVER = [
        User::AUTH_TYPE_MICROSOFT => [
            self::POP3_PROTOCOL => 'outlook.office365.com',
            self::IMAP_PROTOCOL => 'outlook.office365.com',
        ],
        User::AUTH_TYPE_GOOGLE => [
            self::POP3_PROTOCOL => 'pop.gmail.com',
            self::IMAP_PROTOCOL => 'imap.gmail.com',
        ],
    ];

    private $_companyGlobalConfig = null;

    private $_contexts = null;

    private $_profileExitcodeDescription = array(
        1  => 'There was no mail awaiting retrieval.
            (There may have been old mail still on the server but not selected for retrieval.)',
        2  => "An error was encountered when attempting to open a socket to retrieve mail.
            If you don't know  what  a  socket  is,  don't worry  about  it -- just treat this as an 'unrecoverable error'.
            This error can also be because a protocol fetchmail wants to use is not listed in /etc/services.",
        3  => 'The user authentication step failed.  This usually means that a bad user-id, password, or APOP id was specified.
            Or it may mean  that  you  tried to run fetchmail under circumstances where it did not have standard input
            attached to a terminal and could not prompt for a missing password.',
        4  => 'Some sort of fatal protocol error was detected',
        5  => 'There was a syntax error in the arguments to fetchmail',
        6  => 'The run control file had bad permissions.',
        7  => 'There was an error condition reported by the server.  Can also fire if fetchmail timed out while waiting for the server.',
        8  => 'Client-side exclusion error.  This means fetchmail either found another copy of itself already running,
            or failed in such away that it is not sure whether another copy is running.',
        9  => 'The  user  authentication step failed because the server responded "lock busy".  Try again after a brief pause!
            This error is not implemented for all protocols, nor for all servers.
            If not implemented  for  your  server,  "3"  will  be  returned instead,  see  above.
            May  be returned when talking to qpopper or other servers that can respond with "lock busy" or some similar text containing the word "lock".',
        10 => 'The fetchmail run failed while trying to do an SMTP port open or transaction.',
        11 => 'Fatal DNS error.  Fetchmail encountered an error while performing a DNS lookup at startup and could not proceed.',
        12 => 'BSMTP batch file could not be opened.',
        13 => 'Poll terminated by a fetch limit (see the --fetchlimit option).',
        14 => 'Server busy indication.',
        23 => 'Internal error.  You should see a message on standard error with details.',
        24 => 'These are internal codes and should not appear externally.',
    );

    /**
     * FaxSmsProfiles constructor.
     *
     * @param \Wildix\CompanyGlobalConfig        $companyGlobalConfig Company global config.
     * @param \Wildix\Dialplan\ContextRepository $contextRepository   Context repository.
     */
    public function __construct(
        \Wildix\CompanyGlobalConfig $companyGlobalConfig,
        \Wildix\Dialplan\ContextRepository $contextRepository
    ) {
        $this->_companyGlobalConfig = $companyGlobalConfig;
        $this->_contexts            = $contextRepository;
    }

    private function getCompanyGlobalConfig() {
        return $this->_companyGlobalConfig;
    }

    public function getContexts() {
        return $this->_contexts;
    }

    private function getMailToFaxOptions($profileId = 0) {
        $return = [
            'server'       => '',
            'username'     => '',
            'password'     => '',
            'ssl'          => 'off',
            'protocol'     => self::POP3_PROTOCOL,
            'protection'   => '',
            'authType'     => 'Basic',
            'authProvider' => '',
        ];

        $filename = self::COMPANIES_DIRECTORY . '/0/fetchmailrc_' . $profileId;
        if (!file_exists($filename) || !is_readable($filename)) {
            return $return;
        }

        $regex       = "/.*poll\s'?([^']*)'?\swith\sproto\s(pop3|imap)?(:|\sport\s([^:]+):)?\n\s+(user|auth oauthbearer username)\s'(.*)'\s+(passwordfile|there\swith\spassword)\s'(.*)'\s+is.*/is";
        $content     = file_get_contents($filename);
        $matchResult = preg_match($regex, $content, $matches);
        if ($matchResult) {
            $return['server']   = $matches[1];
            $return['protocol'] = $matches[2];

            if (strpos($content, ' keep') !== false) {
                $return['keep'] = 'on';
            } else {
                $return['keep'] = 'off';
            }

            $profileRegexp = '/^#\sprofile:\s([^\n]*)/';
            if (preg_match($profileRegexp, $content, $m)) {
                $return['profileName'] = $m[1];
            }

            $protectionRegexp = '/#\sprotection:\s([^\n]*)/';
            if (preg_match($protectionRegexp, $content, $m)) {
                $return['protection'] = $m[1];
            }

            $provider = '/#\sprovider:\s([^\n]*)/';
            if (preg_match($provider, $content, $m)) {
                $return['authProvider'] = $m[1];
            }

            $ssl = stripos($content, 'sslmode wrapped');
            if ($ssl) {
                $return['ssl'] = 'on';
            }

            if (stripos($content, 'auth oauthbearer username')) {
                $return['authType'] = 'OAuth2';

                $return['oAuthEmail'] = '';

                $accessTokenFile = sprintf(self::OAUTH_CONFIG_ACCESS_FILE_PATTERN, $profileId);
                if (file_exists($accessTokenFile) && is_readable($accessTokenFile)) {
                    $return['oAuthEmail'] = $matches[6];
                }
            } else {
                $return['authType'] = 'Basic';
                $return['username'] = $matches[6];
                $return['password'] = $matches[8];
            }
        }
        return $return;
    }

    public function getProfiles() {
        $profiles = array();

        $files = glob(self::COMPANIES_DIRECTORY . '0/fetchmailrc_*');
        if (is_array($files)) {
            foreach ($files as $file) {
                if (preg_match('/_(\d+)/', $file, $matches)) {
                    $profile = $this->getProfile($matches[1]);

                    if (!isset($profile['profileName'])) {
                        continue;
                    }

                    $profile['keep'] = $profile['keep'] === 'on';
                    $profile['ssl']  = $profile['ssl'] === 'on';

                    $profiles[] = $profile;
                }
            }
        }

        // This is the format for WMS.DataTables (we can also pass 'rowsPerPage', 'total' etc.)
        return array('records' => $profiles);
    }

    private function getProfileId($profileId) {
        if ($profileId == 'new') {
            $output = array();

            $files = glob(self::COMPANIES_DIRECTORY . '0/fetchmailrc_*');
            if (is_array($files)) {
                foreach ($files as $file) {
                    if (preg_match('/_(\d+)/', $file, $matches)) {
                        $output[] = $matches[1];
                    }
                }
            }

            $profileId = count($output);

            $len = count($output);
            for ($i = 0; $i < $len; $i++) {
                if (!in_array($i, $output)) {
                    $profileId = $i;
                    break;
                }
            }
        }

        return $profileId;
    }

    private function setMailToFaxOptions($options) {
        $filename = self::COMPANIES_DIRECTORY . '/0/fetchmailrc_' . $options['profileId'];

        $options['keep'] = $options['keep'] == 'on' ? 'keep' : '';

        if (empty($options['oAuthEmail'])) {
            $this->removeOauthConfigFiles((int)$options['profileId']);
        }

        if ($this->isOauth($options['authType'])) {
            $options['ssl']    = 'on';
            $options['server'] = $this->getProviderServer($options['authProvider'], $options['protocol']);
        }

        if ($options['server'] == '') {
            $output = '';
        } else {
            $port = '';
            if ($options['protocol'] === self::IMAP_PROTOCOL) {
                $port = $options['ssl'] == 'on' ? 993 : 143;
            } else if ($options['protocol'] === self::POP3_PROTOCOL) {
                $port = $options['ssl'] == 'on' ? 995 : 110;
            }

            $output  = '# profile: ' . $options['profileName'];
            $output .= ($options['protection'] != '' ? "\n" . '# protection: ' . $options['protection'] : '');
            $output .= ($this->isOauth($options['authType']) && $options['authProvider'] != '' ? "\n" . '# provider: ' . $options['authProvider'] : '');

            $output .= "\npoll '" . $options['server'] . "' with proto " . $options['protocol'] . ' port ' . $port;

            if ($this->isOauth($options['authType'])) {
                $this->writeOauthConfigFiles($options);

                $output .= ":\n\tauth oauthbearer username '" . ($options['oAuthEmail'] ?? '') . "'\n\t";
                $output .= "passwordfile '/rw2/etc/companies.d/0/fetchmail_access_" . $options['profileId'] . "'\n\t";
                $output .= "is 'wms' here options " . $options['keep'] . ' sslmode wrapped' . "\n";
            } else {
                $output .= ":\n\tuser '" . $options['username'] . "' there with password '" . $options['password'];
                $output .= "' is 'wms' here options " . $options['keep'] . ($options['ssl'] == 'on' ? ' sslmode wrapped' : '') . "\n";
                $this->removeOauthConfigFiles((int)$options['profileId']);
            }
        }

        $oldContent = @file_get_contents($filename);
        if ($oldContent === $output) {
            return;
        }

        if (!file_put_contents($filename, $output)) {
            throw new Exception(Error::getMessage(Error::IsNotWriteable) . ' ' . $filename);
        }

        chmod($filename, 0600);
    }

    private function validateProfile($params) {
        $form = new Validator(
            $params,
            [
                'profileName',
                'server',
                'LOCALHEADERINFO',
                'LOCALSTATIONID',
                'protocol',
                'keep',
                'ssl',
                'faxContext',
                'authType',
                'authProvider',
                'oAuthEmail',
            ]
        );

        $requiredFields = [
            'profileName',
            'protocol',
            'authType',
        ];

        if ($this->isOauth($params['authType'])) {
            $requiredFields[] = 'authProvider';

            $form->rule(
                'in',
                ['authProvider'],
                [User::AUTH_TYPE_GOOGLE, User::AUTH_TYPE_MICROSOFT]
            );
        } else {
            $requiredFields[] = 'server';
        }

        $form->rule(
            'required',
            $requiredFields
        );

        $form->rule(
            'alphaNum',
            ['profileName']
        );

        $form->rule(
            'regex',
            ['LOCALHEADERINFO', 'LOCALSTATIONID'],
            '/^\s*[\-+_0-9a-zA-Z]*\s*$/'
        );

        $form->rule(
            'in',
            ['protocol'],
            [self::POP3_PROTOCOL, self::IMAP_PROTOCOL]
        );

        $form->rule(
            'boolean',
            ['keep', 'ssl']
        );

        $contexts = $this->getContexts()->getNames();
        $form->rule(
            'in',
            ['faxContext'],
            array_keys($contexts)
        );

        if (!$form->validate()) {
            throw new \Wildix\InvalidDataException('Validation error', $form->errors());
        }
    }

    public function saveProfile($datastore) {
        $this->validateProfile($datastore);

        $datastore['profileId'] = $this->getProfileId($datastore['profileId']);

        $this->setMailToFaxOptions($datastore);

        $configVariables = array(
            'GFAXCONTEXT' . $datastore['profileId'] => $datastore['faxContext'],
            'LOCALSTATIONID' . $datastore['profileId'] => $datastore['LOCALSTATIONID'],
            'LOCALHEADERINFO' . $datastore['profileId'] => $datastore['LOCALHEADERINFO'],
        );

        $this->getCompanyGlobalConfig()->setVariables($configVariables);

        \Wildix\CallweaverManager::getInstance()->reloadModule('pbx_config');

        return $this->getProfile($datastore['profileId']);
    }

    public function getProfile($profileId) {
        $mOptions   = $this->getMailToFaxOptions($profileId);
        $faxContext = $this->getCompanyGlobalConfig()->getVariable('GFAXCONTEXT' . $profileId);
        $faxContext = $faxContext ? $faxContext : 'users';

        $mOptions['LOCALSTATIONID']  = $this->getCompanyGlobalConfig()->getVariable('LOCALSTATIONID' . $profileId);
        $mOptions['LOCALHEADERINFO'] = $this->getCompanyGlobalConfig()->getVariable('LOCALHEADERINFO' . $profileId);
        $mOptions['faxContext']      = $faxContext;
        $mOptions['profileId']       = $profileId;

        return $mOptions;
    }

    public function removeProfiles($ids) {
        foreach ($ids as $profileId) {
            $file = sprintf('%s/0/fetchmailrc_%d', self::COMPANIES_DIRECTORY, $profileId);
            if (file_exists($file)) {
                unlink($file);
            }
            $this->removeOauthConfigFiles((int)$profileId);
        }
    }

    public function runFetchmailTest($profileId) {
        $resultTestOutFilename = self::PROFILE_TEST_DIRECTORY . 'profile_test_' . $profileId . '_out';
        $resultTestErrFilename = self::PROFILE_TEST_DIRECTORY . 'profile_test_' . $profileId . '_err';

        $profileFile = '/rw2/etc/companies.d/0/fetchmailrc_' . $profileId;
        if (!file_exists($profileFile)) {
            return array('queued' => false);
        }

        $testProfileBashScriptFileName = self::PROFILE_TEST_DIRECTORY . 'test_fetchmailrc_' . $profileId . '.sh';

        $testProfileBashScriptContent  = '#!/bin/bash' . "\n";
        $testProfileBashScriptContent .= 'set +e' . "\n";
        $testProfileBashScriptContent .= 'fetchmail --timeout=10 --pidfile /var/www/fetchmail.pid -i /rw2/etc/fetchids ';
        $testProfileBashScriptContent .= '-f /rw2/etc/companies.d/0/fetchmailrc_' . $profileId . ' -k -s --verbose --uidl -c ';
        $testProfileBashScriptContent .= '>>' . $resultTestOutFilename . ' 2>>' . $resultTestErrFilename . "\n";
        $testProfileBashScriptContent .= 'echo exitcode=$? >>' . $resultTestOutFilename . "\n";
        $testProfileBashScriptContent .= 'set -e' . "\n";

        file_put_contents($testProfileBashScriptFileName, $testProfileBashScriptContent);
        chmod($testProfileBashScriptFileName, 0755);

        $taskCmd = 'nohup bash -c "' . $testProfileBashScriptFileName . ' &" >' . $resultTestOutFilename;

        exec($taskCmd, $output, $result);

        return array(
            'queued'    => $result === 0,
            'profileId' => $profileId,
        );
    }

    public function getFetchmailTestStatus($profileId) {
        $testProfileBashScriptFileName = self::PROFILE_TEST_DIRECTORY . 'test_fetchmailrc_' . $profileId . '.sh';

        $resultTestOutFilename = self::PROFILE_TEST_DIRECTORY . 'profile_test_' . $profileId . '_out';
        $resultTestErrFilename = self::PROFILE_TEST_DIRECTORY . 'profile_test_' . $profileId . '_err';
        if (!file_exists($resultTestOutFilename) || !file_exists($resultTestErrFilename)) {
            return array(
                'queued'    => false,
                'profileId' => $profileId,
            );
        }
        $resultContent = file_get_contents($resultTestOutFilename);
        if (file_exists($resultTestErrFilename)) {
            $resultContent .= file_get_contents($resultTestErrFilename);
        }
        $isEnded = strpos($resultContent, 'exitcode') !== false;
        if ($isEnded) {
            [$resultContent, $code] = explode('exitcode=', trim($resultContent));

            $resultContent .= 'Note: ' . $this->_profileExitcodeDescription[intval($code)] . "\n";

            unlink($resultTestOutFilename);
            unlink($resultTestErrFilename);
            unlink($testProfileBashScriptFileName);
        }

        return array(
            'content'   => $resultContent,
            'queued'    => !$isEnded,
            'profileId' => $profileId,
        );
    }

    public function writeOauthConfigFiles(array $data): void {
        $accessTokenFile  = sprintf(self::OAUTH_CONFIG_ACCESS_FILE_PATTERN, $data['profileId']);
        $refreshTokenFile = sprintf(self::OAUTH_CONFIG_REFRESH_FILE_PATTERN, $data['profileId']);

        if ($data['accessToken'] && !file_put_contents($accessTokenFile, $data['accessToken'])) {
            throw new Exception(Error::getMessage(Error::IsNotWriteable) . ' ' . $accessTokenFile);
        }

        if ($data['refreshToken'] && !file_put_contents($refreshTokenFile, $data['refreshToken'])) {
            throw new Exception(Error::getMessage(Error::IsNotWriteable) . ' ' . $refreshTokenFile);
        }

        chmod($accessTokenFile, 0600);
        chmod($refreshTokenFile, 0600);
    }

    public function removeOauthConfigFiles(int $profileId): void {
        $files = [
            sprintf(self::OAUTH_CONFIG_ACCESS_FILE_PATTERN, $profileId),
            sprintf(self::OAUTH_CONFIG_REFRESH_FILE_PATTERN, $profileId),
        ];

        array_map(
            function ($file) {
                if (file_exists($file)) {
                    unlink($file);
                }
            },
            $files
        );
    }

    private function getProviderServer(string $provider, string $protocol): string {
        return self::OAUTH_PROVIDER_PROTOCOL_SERVER[mb_strtolower($provider)][$protocol];
    }

    private function isOauth(string $authType): bool {
        return mb_strtolower($authType) === User::AUTH_TYPE_OAUTH2;
    }
}
